/* ***** BEGIN LICENSE BLOCK *****
 * 
 * Copyright (c) 1997-2008 Aliasource - Groupe LINAGORA
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  http://www.obm.org/                                              
 * 
 * ***** END LICENSE BLOCK ***** */
const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

function JSObjEventShell(aEvent /* obmIEvent */) {
  aEvent.QueryInterface(Ci.obmIEventRecurrence);
  var obj = {};
  obj.uid = aEvent.uid;
  obj.owner = aEvent.owner;
  obj.priority = aEvent.priority;
  obj.classification = aEvent.classification;
  obj.allday = aEvent.allday;
  obj.date = aEvent.date;
  obj.duration = aEvent.duration;
  obj.title = aEvent.title;
  obj.description = aEvent.description;
  obj.category = aEvent.category;
  obj.location = aEvent.location;

  obj.attendees = [];
  var obmAttendees = aEvent.getAttendees({});
  for (var i = 0; i < obmAttendees.length; i++) {
    var attendee = {};
    attendee.email = obmAttendees[i].email;
    attendee.displayName = obmAttendees[i].displayName;
    attendee.status = obmAttendees[i].status;
    attendee.type = obmAttendees[i].type;
    obj.attendees.push(attendee);
  }
  if (obj.attendees.length == 0) {
    delete(obj.attendees);
  }
  
  obj.recurrence = {};
  obj.recurrence.kind = aEvent.kind;
  obj.recurrence.frequence = aEvent.frequence;
  obj.recurrence.days = aEvent.days;
  obj.recurrence.end = aEvent.end;
  obj.recurrence.exception = aEvent.getExceptions({});
  if (obj.recurrence.exception.length == 0) {
    delete(obj.recurrence.exception);
  }

	if (aEvent.alert != -2) {
		obj.alert = aEvent.alert;
	}
  
  return obj;
}

function ObmCalendarProxy() {
  // Init logger
  this.logger = Cc["@obm.org/calendar/logger;1"].createInstance(Ci.obmILogger);
  var writer = Cc["@obm.org/calendar/writer;1?type=unixfile"].getService(Ci.obmILoggerWriter);
  this.logger.setLevelWriter(this.logger.LOAD_LEVEL, writer);
  this.logger.setHeader("obmCalendarProxy: ");
  this.logger.debug("new obmCalendarProxy()", false, false);

	this.SOAPClient = Cc["@guru4.net/soapclient;1"].getService().wrappedJSObject;

	this._listeners = [];
	this._url = null;
	this._token = {
		user: null,
		domain: null,
		session_id: null
	};
};

ObmCalendarProxy.prototype = {
	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) &&
			!aIID.equals(Ci.obmICalendarProxy) &&
			!aIID.equals(Ci.obmIProxyBase)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},

	/*
	 * obmIProxyBase
	 */
	addListener: function(listener) {
		for (var i = 0; i < this._listeners.length; ++i) {
			if (this._listeners[i] == listener)
				return;
		}
		this._listeners.push(listener);
	},

	removeListener: function(listener) {
		this._listeners = [];
	},
	_notifyListeners: function(aResultCode, aMethodName, aObj, aData) {
		// A copy is needed, a new listeners could has been added during the
		// callbacks
		var listeners = this._listeners.slice(0, this._listeners.length);
		for (var i = 0; i < listeners.length; ++i) {
			try {
				listeners[i].onResponse(aResultCode, aMethodName, aObj, aData);
			} catch(e) {
				// Continue to notify listeners
        this.logger.error(e + " (ignored)");
			}
		}
	},

	get token() {
		return this._token;
	},
	set token(aToken) {
		this._token = aToken;
	},

	get WSDLUrl() {
		return this._url;
	},
	set WSDLUrl(aURL) {
		this._url = aURL;
	},



	/*
	 * obmICalendarProxy
	 */
	_callback: function(aMethodName, aObj, aXml, aData) {
    this.logger.debug("_callback(aMethodName = " + aMethodName + ")", false, false);
		if (aObj && aObj.name == "Error") {
			var msg = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
			msg.data = aObj.fileName;
			this._notifyListeners(Cr.NS_ERROR_FAILURE, aMethodName, msg, aData);
			return;
		}
		switch(aMethodName) {
			case "logUserIn":
				if (!aObj.token || !aObj.token[0]) {
				  var msg = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
				  msg.data = "No token available";
				  this._notifyListeners(Cr.NS_ERROR_FAILURE, aMethodName, msg, aData);
				  return;
				}
				this._token.user = aObj.token[0].user[0];
				this._token.domain = aObj.token[0].domain[0];
				this._token.session_id = aObj.token[0].session_id[0];
				this._notifyListeners(Cr.NS_OK, aMethodName, null, aData);
				break;
			case "getSync":
				this._notifyListeners(Cr.NS_OK, aMethodName, new GetSyncEventResponseShell(aObj._return[0]), aData);
				break;
			case "createEvent":
        var id = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
        id.data = aObj._return[0];
				this._notifyListeners(Cr.NS_OK, aMethodName, id, aData);
				break;
			case "modifyEvent":
				this._notifyListeners(Cr.NS_OK, aMethodName, new EventShell(aObj._return[0]), aData);
				break;
			case "removeEvent":
				this._notifyListeners(Cr.NS_OK, aMethodName, null, aData);
				break;
			case "listCalendars":
				var XPArray = {
					list: aObj._return[0]._return,
					QueryInterface: function(aIID) {
						if (!aIID.equals(Ci.nsISupports) &&
							!aIID.equals(Ci.nsIArray)) {
							throw Cr.NS_ERROR_NO_INTERFACE;
						}
						return this;
					},
					get length() {return this.list.length;},
					enumerate: function() {throw Cr.NS_ERROR_NOT_IMPLEMENTED},
					indexOf: function(idx, elt) {throw Cr.NS_ERROR_NOT_IMPLEMENTED},
					queryElementAt: function(idx, uuid) {
						var cal = new CalendarShell(this.list[idx]);
						return cal.QueryInterface(uuid);
					}
				};
				this._notifyListeners(Cr.NS_OK, aMethodName, XPArray, aData);
				break;
			default:
      this._notifyListeners(Cr.NS_ERROR_FAILURE, aMethodName, null, aData);
		};
	},

	getSync: function(aCalName, aLastSync, aData) {
    this.logger.debug("getSync()", false, false);
		var p = this.SOAPClient.newSOAPClientParameters();
		p.add("token", this._token);
		p.add("book", aCalName);
		p.add("lastsync", aLastSync);
		// closure
		var self = this;
		this.SOAPClient.invoke(this._url, "getSync", p, true, function(name, obj, xml, data) {self._callback(name, obj, xml, data);}, aData);
	},
	createEvent: function(aCalendarId, aEvent, aData) {
    this.logger.debug("createEvent()", false, false);
		var p = this.SOAPClient.newSOAPClientParameters();
		p.add("token", this._token);
		p.add("calendar", aCalendarId);
		p.add("evt", JSObjEventShell(aEvent));
		// closure
		var self = this;
		this.SOAPClient.invoke(this._url, "createEvent", p, true, function(name, obj, xml, data) {self._callback(name, obj, xml, data);}, aData);
	},
	modifyEvent: function(aCalendarId, aEvent, aUpdateAttendees, aData) {
    this.logger.debug("modifyEvent()", false, false);
		var p = this.SOAPClient.newSOAPClientParameters();
		p.add("token", this._token);
		p.add("calendar", aCalendarId);
		p.add("evt", JSObjEventShell(aEvent));
    p.add("updateAttendees", aUpdateAttendees);
		// closure
		var self = this;
		this.SOAPClient.invoke(this._url, "modifyEvent", p, true, function(name, obj, xml, data) {self._callback(name, obj, xml, data);}, aData);
	},
	removeEvent: function(aCalendarId, aUid, aData) {
    this.logger.debug("removeEvent()", false, false);
		var p = this.SOAPClient.newSOAPClientParameters();
		p.add("token", this._token);
		p.add("calendar", aCalendarId);
		p.add("id", aUid);
		// closure
		var self = this;
		this.SOAPClient.invoke(this._url, "removeEvent", p, true, function(name, obj, xml, data) {self._callback(name, obj, xml, data);}, aData);
	},
	logUserIn: function(aLogin, aPwd, aData) {
    this.logger.debug("logUserIn()", false, false);
		var p = this.SOAPClient.newSOAPClientParameters();
		var p = this.SOAPClient.newSOAPClientParameters();
		p.add("login", aLogin);
		p.add("password", aPwd);
		// closure
		var self = this;
		this.SOAPClient.invoke(this._url, "logUserIn", p, true, function(name, obj, xml, data) {self._callback(name, obj, xml, data);}, aData);
	},
	listCalendars: function(aData) {
    this.logger.debug("listCalendars()", false, false);
		var p = this.SOAPClient.newSOAPClientParameters();
		p.add("token", this._token);
		// closure
		var self = this;
		this.SOAPClient.invoke(this._url, "listCalendars", p, true, function(name, obj, xml, data) {self._callback(name, obj, xml, data);}, aData);
	},

};

function CalendarShell(inner) {
	this._mInner = inner;
}

CalendarShell.prototype = {
	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) &&
			!aIID.equals(Ci.obmICalendarInfo)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},
	// obmICalendarInfo
	get uid() {return this._mInner.uid[0]},
	get firstname() {return this._mInner.firstname[0]},
	get lastname() {return this._mInner.lastname[0]},
	get mail() {return this._mInner.mail[0]},
	get read() {return this._mInner.read[0] == "true"},
	get write() {return this._mInner.write[0] == "true"},
};


function EventShell(inner) {
	this._mInner = inner;
}

EventShell.prototype = {
	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) &&
			!aIID.equals(Ci.obmIEventRecurrence) &&
			!aIID.equals(Ci.obmIEvent)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},
	// obmIEvent
	get uid() {return this._mInner.uid},
	get owner() {return this._mInner.owner},
	get priority() {return this._mInner.priority},
	get classification() {return this._mInner.classification},
	get allday() {return this._mInner.allday == "true"},
	get date() {return this._mInner.date},
	get duration() {return this._mInner.duration},
	get title() {return this._mInner.title},
	get description() {return this._mInner.description},
	get category() {return this._mInner.category},
	get location() {return this._mInner.location},
	get recurrence() {return this;},
	getAttendees: function(count) {
    var attendees = [];
    if (this._mInner.attendees)
      attendees = this._mInner.attendees;
		count.value = attendees.length;
		return attendees;
	},
	// obmIEventRecurrence
  get kind() {return this._mInner.recurrence[0].kind},
	get frequence() {return this._mInner.recurrence[0].frequence},
	get days() {return this._mInner.recurrence[0].days},
	get end() {return this._mInner.recurrence[0].end},
	getExceptions: function(count) {
		var recurrence = this._mInner.recurrence[0];
		if (!("exception" in recurrence))
			return;
		count.value = recurrence.exception.length;
		return(recurrence.exception);
	},
	get alert() {return this._mInner.alert}
};

function GetSyncEventResponseShell(inner) {
	this._mInner = inner;
	this._updated = [];
	this._removed = [];

	if (inner.updated) for (var i = 0; i < inner.updated.length; i++) {
		this._updated.push(new EventShell(inner.updated[i]));
	}
	if (inner.removed) for (var i = 0; i < inner.removed.length; i++) {
		this._removed.push(inner.removed[i]);
	}
}

GetSyncEventResponseShell.prototype = {
	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) &&
			!aIID.equals(Ci.obmICalendarSyncResponse)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},
	getUpdatedEvents: function(count) {
		count.value = this._updated.length;
		return this._updated;
	},
	getRemovedEventIds: function(count) {
		count.value = this._removed.length;
		return this._removed;
	},
	get lastsync() {
		return this._mInner.lastsync[0];
	}
};


// module

var ObmCalendarProxyModule = {
	mCID: Components.ID("{7aff35b3-4754-40f2-b98a-4453b1587e5d}"),
	mContractID: "@obm.org/obm-sync/calendarproxy;1",
	registerSelf: function (compMgr, fileSpec, location, type) {
		compMgr = compMgr.QueryInterface(Ci.nsIComponentRegistrar);
		compMgr.registerFactoryLocation(
			this.mCID,
			"OBM Calendar Proxy",
			this.mContractID,
			fileSpec,
			location,
			type
		);
	},
	getClassObject: function (compMgr, cid, iid) {
		if (!cid.equals(this.mCID)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		if (!iid.equals(Ci.nsIFactory)) {
			throw Cr.NS_ERROR_NOT_IMPLEMENTED;
		}
		return this.mFactory;
	},
	mFactory: {
		createInstance: function (outer, iid) {
			if (outer != null) {
				throw Cr.NS_ERROR_NO_AGGREGATION;
			}
			return (new ObmCalendarProxy()).QueryInterface(iid);
		}
	},

	canUnload: function(compMgr) {
		return true;
	}
};

function NSGetModule(compMgr, fileSpec) {
	return ObmCalendarProxyModule;
}

