/* ***** BEGIN LICENSE BLOCK *****
 * 
 * Copyright (c) 1997-2008 Aliasource - Groupe LINAGORA
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  http://www.obm.org/                                              
 * 
 * ***** END LICENSE BLOCK ***** */
const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

function ObmCategoriesSync() {
  //Init logger
  this.logger = Cc["@obm.org/calendar/logger;1"].createInstance(Ci.obmILogger);
  var writer = Cc["@obm.org/calendar/writer;1?type=unixfile"].getService(Ci.obmILoggerWriter);
  this.stringBundle = Cc["@mozilla.org/intl/stringbundle;1"].getService(Ci.nsIStringBundleService)
     	.createBundle("chrome://obm-extension/locale/calendar/errors_warnings.properties");
  this.logger.setLevelWriter(this.logger.LOAD_LEVEL, writer);
  this.logger.setHeader("obmCategoriesSync: ");
  this.logger.debug("new obmCategoriesSync()", false, false);

  this.criticalMethods = ["listCategories"];
};

function Context() {
	this.wrappedJSObject = this;
};


ObmCategoriesSync.prototype = {
	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) &&
			!aIID.equals(Ci.obmICategoriesSync)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},


	_getLocalizedMessage: function(aString) {
		dump(""+aString+"\n");
		dump("obm.syncerror."+aString+"\n");
		try {
			return this.stringBundle.GetStringFromName("obm.syncerror."+aString);
		} catch(e) {
			return aString;
		}
	},
	_isCriticalMethod: function(aMethodeName) {
		for each (method in this.criticalMethods) {
			if (aMethodeName == method) {
				return true;
			}
		}
		return false;
	},
	_printableError: function(aMethodName, aMsg) {
		var printError = "srvError"; 
		
		if (aMsg == "No token available") {
			printError = "authFault";
		}
		
		if (aMsg == "Connection failure") {
			printError = "connectFail";
		}
		
		return this._getLocalizedMessage(printError);
	},
	_getOBMCredentials: function(aLogin, aPassword) {

		if ( !UtilsOBM.getOBMCredentials(aLogin, aPassword) ||
				!aLogin.value ||
				aLogin.value == "" ||
				!aPassword.value ||
				aPassword.value == "") {
			var error = "NO_LOGIN_INFO";
      this.logger.error(error, true, false);
			this.callerListener.onError(error);
			throw(error);
		}
	},

	_ensureValidResponse: function(aResCode, aMethodName, aResult, aValidMethodName) {
		if (aResCode != Cr.NS_OK) {
			var msgError = "unknown";
			if (aResult instanceof Ci.nsISupportsString)
				msgError = aResult;
			var printError = this._printableError(aMethodName, msgError);
      msgError = aMethodName + ": Reported error: " + msgError + "(" + aResCode + ")";
      this.logger.error(msgError, true, false);
			if (this.callerListener) {
				if (this._isCriticalMethod(aMethodName)) {
					this.callerListener.onStop(aResCode, printError);
				} else {
					this.callerListener.onError(printError);
				}
			}
			return false;
		}

		if (aMethodName != aValidMethodName) {
      var msgError = "Method " + aMethodName + " unexpected. Expected: " + aValidMethodName;
      this.logger.error(msgError, true, false);
			this.callerListener.onError(msgError);
			return false;
		}
		return true;
	},

  _onListCategories: function(aResCode, aMethodName, aResult) {
    this.logger.debug("_onListCategories()", false, false);
	this.callerListener.onOperationDone(true, null);
	this.logger.debug("_onListCategories", false, false);
	this.CatProxy.removeListener();
	if (!this._ensureValidResponse(aResCode, aMethodName, aResult, "listCategories")) {
		return;
	};

	var currentCategories = getLocalizedPref("calendar.categories.names");
	if (currentCategories && currentCategories != "") {
		currentCategories = currentCategories.split(",");
	} else {
		currentCategories = [];
	}
	
	var newCategories = [];
	for each (var category in currentCategories) {
		if ( !category.match("[OBM]")) {
			newCategories.push(category);
		}
	}

	var serverCategories = aResult.QueryInterface(Ci.obmICategoriesResponse).getCategories({});
	for each(var category in serverCategories) {
		var label = "[OBM] " + category.label;
		newCategories.push(label);
	}
	
	setLocalizedPref("calendar.categories.names", currentCategories.join(","));
	this.callerListener.onStop(Cr.NS_OK, "done");
  },

	/*
	 * obmIAddressBookSync
	 */

	mergeCategories: function(aListener) {
    this.logger.debug("mergeCategories()", false, false);
		this.logger.debug("mergeCategories", false, false);
		this.callerListener = aListener;
		this.callerListener.onStart();

		try {
			var WSDLUrl = UtilsOBM.getCharPref("extensions.obm.server") + "/Categories";
		} catch(e) {
			var error = "NO_SERVER_INFO";
      this.logger.error(error, true, false);
			this.callerListener.onError(error);
			throw(error);
		}


		this.CatProxy = Cc["@obm.org/obm-sync/categoriesproxy;1"].createInstance(Ci.obmICategoriesProxy);
		this.CatProxy.WSDLUrl = WSDLUrl;

		this.callerListener.onNewOperation("listCategories", {});
		var self = this;
		this.CatProxy.addListener(function(aResCode, aMethodName, aResult, aData) {
			self._onListCategories(aResCode, aMethodName, aResult)
		});
    this.CatProxy.listCategories();

    return;
	}
};

// moduleo

var ObmCategoriesSyncModule = {
	mCID: Components.ID("{a2fb53d3-5409-425c-ab86-a64e285f7cc0}"),
	mContractID: "@obm.org/obm-sync/categoriessync;1",
	registerSelf: function (compMgr, fileSpec, location, type) {
		compMgr = compMgr.QueryInterface(Ci.nsIComponentRegistrar);
		compMgr.registerFactoryLocation(
			this.mCID,
			"OBM Categories Sync",
			this.mContractID,
			fileSpec,
			location,
			type
		);
	},
	getClassObject: function (compMgr, cid, iid) {

		if (!cid.equals(this.mCID)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		if (!iid.equals(Ci.nsIFactory)) {
			throw Cr.NS_ERROR_NOT_IMPLEMENTED;
		}
		if (!this._utilsLoaded) {
			var scriptLoader = Cc["@mozilla.org/moz/jssubscript-loader;1"].createInstance(Ci.mozIJSSubScriptLoader);
			scriptLoader.loadSubScript("chrome://obm-extension/content/calendar/UtilsOBM.js");
      scriptLoader.loadSubScript("chrome://calendar/content/calUtils.js");
			this._utilsLoaded = true;
		}

		return this.mFactory;
	},
	mFactory: {
		createInstance: function (outer, iid) {
			if (outer != null) {
				throw Cr.NS_ERROR_NO_AGGREGATION;
			}
			return (new ObmCategoriesSync()).QueryInterface(iid);
		}
	},

	canUnload: function(compMgr) {
		return true;
	}
};

function NSGetModule(compMgr, fileSpec) {
	return ObmCategoriesSyncModule;
}


