/* ***** BEGIN LICENSE BLOCK *****
 * 
 * Copyright (c) 1997-2008 Aliasource - Groupe LINAGORA
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  http://www.obm.org/                                              
 * 
 * ***** END LICENSE BLOCK ***** */
const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

const RDFService = Cc["@mozilla.org/rdf/rdf-service;1"].getService(Ci.nsIRDFService);

function wrapCallbackForErrorHandling(callback, errorHandler) {
    return (function(aResCode, aMethodName, aResult, aData) {
	try {
	    callback(aResCode, aMethodName, aResult, aData);
	} catch(e) {
	    errorHandler(aMethodName, e);
	}
    });
}

function ObmRAddressBookSync() {
	// Init logger
	this.logger = Cc["@obm.org/calendar/logger;1"].createInstance(Ci.obmILogger);
	var writer = Cc["@obm.org/calendar/writer;1?type=unixfile"].getService(Ci.obmILoggerWriter);
	this.stringBundle = Cc["@mozilla.org/intl/stringbundle;1"].getService(Ci.nsIStringBundleService).createBundle("chrome://obm-extension/locale/calendar/errors_warnings.properties");
	this.logger.setLevelWriter(this.logger.LOAD_LEVEL, writer);
	this.logger.setHeader("obmRAddressBookSync.js: ");
	this.logger.debug("new obmRAddressBookSync()", false, false);

	this.criticalMethods = ["doLogin", "getAddressBookSync"];

	this._localDiff = {};
	this._remoteDiff = {};

	this.abVersion = Cc["@mozilla.org/abmanager;1"] ? 3 : 2;

	this.obmAbManager = Cc["@obm.org/addressbooksmanager;1"].getService(Ci.obmIAddressBooksManager);
};

function Context() {
	this.wrappedJSObject = this;
};

ObmRAddressBookSync.prototype = {
	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) && !aIID.equals(Ci.obmIRAddressBookSync)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},

	/*
	 * Private
	 */
	_onError: function(aMethodName, error) {
	    if (this._isCriticalMethod(aMethodName)) {
		this.callerListener.onStop(Cr.NS_ERROR_FAILURE, error);
	    } else {
		this.callerListener.onError(error);
	    }
	},

	_addListener: function(callback) {
	    var self = this;
	    self.ABProxy.addListener(wrapCallbackForErrorHandling(callback, function(aMethodName, error) {self._onError(aMethodName, error)}));
	},

	/*
	 * Helpers
	 */
	_getLocalizedMessage: function(aString) {
		dump("" + aString + "\n");
		try {
			return this.stringBundle.GetStringFromName("obm.syncerror." + aString);
		} catch(e) {
			return aString;
		}
	},
	_isCriticalMethod: function(aMethodeName) {
		for each(method in this.criticalMethods) {
			if (aMethodeName == method) {
				return true;
			}
		}
		return false;
	},
	_printableError: function(aMethodName, aMsg) {
		var printError = "srvError";

		if (aMsg == "No token available") {
			printError = "authFault";
		}

		if (aMsg == "Connection failure") {
			printError = "connectFail";
		}

		return this._getLocalizedMessage(printError);
	},
	_getOBMCredentials: function(aLogin, aPassword) {

		if (!UtilsOBM.getOBMCredentials(aLogin, aPassword) || !aLogin.value || aLogin.value == "" || !aPassword.value || aPassword.value == "") {
			var error = "NO_LOGIN_INFO";
			this.logger.error(error, true, false);
			this.callerListener.onError(error);
			throw (error);
		}
	},
	_getXPStringFromJSString: function(str) {
		var XPString = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
		XPString.data = str;
		return XPString;
	},
	_getJSStringFromXPString: function(XPString) {
		XPString.QueryInterface(Ci.nsISupportsString);
		return XPString.data;
	},

	// DEPRECATED. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager::getAddressBookUids instead
	_getAddressBookNames: function() {
		var names = UtilsOBM.getCharPref("extensions.obm.addressbooks", null);
		return names ? names.split(",") : [];
	},

	_addAddressBookNames: function(newName) {
		// TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager::getAddressBookUids instead
		var names = this._getAddressBookNames();
		for(var i=0; i<names.length; i++) {
			if (names[i] == newName) {
				return;
			}
		}
		var value = "";
		if (names.length == 0) {
			value = newName;
		} else {
			value = UtilsOBM.getCharPref("extensions.obm.addressbooks", null) + "," + newName;
		}
		UtilsOBM.setCharPref("extensions.obm.addressbooks", value);
	},

	_getDatabaseFromABName: function(abName) {
		var abURI = "moz-abmdbdirectory://alia-obm-" + abName + ".mab";
		this.logger.debug("_getDatabaseFromABName: abName=" + abName + ", abURI=" + abURI, false, false);

		if ("@mozilla.org/abmanager;1" in Cc) {
			return "v3";
		} else {
			var AbService = Cc["@mozilla.org/addressbook;1"].getService(Ci.nsIAddressBook);
			var addrDb = AbService.getAbDatabaseFromURI(abURI);
		}

		return addrDb;
	},
	_getDirectoryFromABName: function(abName) {
		var abURI = "moz-abmdbdirectory://alia-obm-" + abName + ".mab";
		if (this.abVersion == 3) {
			var addressBookManager = Components.classes["@mozilla.org/abmanager;1"].getService(Components.interfaces.nsIAbManager);
			var dirPrefId = "ldap_2.servers." + abName; //UtilsOBM.getCharPref("extensions.obm.addressbooks." + abName);
			var iter = addressBookManager.directories;
			while (iter.hasMoreElements()) {
				var data = iter.getNext();
				if (data instanceof Ci.nsIAbDirectory) {
					if (data.dirPrefId == dirPrefId) {
						var directory = data;
					}
				}
			}
		} else {
			var directory = RDFService.GetResource(abURI).QueryInterface(Ci.nsIAbDirectory);
		}
		return directory;
	},
	// DEPRECATED. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager
	_getLocalContact: function(abName, aId) {
		var directory = this._getDirectoryFromABName(abName); //RDFService.GetResource(abURI).QueryInterface(Ci.nsIAbDirectory);
		if (this.abVersion != 3) {
			var addrDb = this._getDatabaseFromABName(abName);
			var aObjectItem = addrDb.getCardFromAttribute(directory, "AnniversaryDay", aId, false);
		} else {
			var aObjectItem = directory.getCardFromProperty("ObmId", aId, false);
		}
		if (!aObjectItem) {
			throw ("Unexpected id (in addressbook " + abName + "): " + aId);
		}
		return aObjectItem;
	},

	_getLastSync: function() {
		this.logger.debug("_getLastSync()", false, false);
		var lastSync = UtilsOBM.getCharPref("extensions.obm.changes." + Ci.nsIAbDirectory + "." + "lastSync", "1970-01-01T00:00:00Z");
		// TODO : need an isValidDate func ?
		if (lastSync == "1970-01-01T00:00:00Z") {
			lastSync = "0";
		}
		return lastSync;
	},
	_setLastSync: function(lastSync) {
		this.logger.debug("SETLASTSYNC (" + lastSync + ") : " + "extensions.obm.changes." + Ci.nsIAbDirectory + "." + "lastSync", false, false);
		UtilsOBM.setCharPref("extensions.obm.changes." + Ci.nsIAbDirectory + "." + "lastSync", lastSync);
	},
	_createContactObjFromId: function(abName, aId, aWillBeCreated) {
		// TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to get a local contact
		var aObjectItem = this._getLocalContact(abName, aId);
		var contact = {};
		contact.QueryInterface = function(aIID) {
			if (!aIID.equals(Ci.nsISupports) && !aIID.equals(Ci.obmIRContact)) {
				throw Cr.NS_ERROR_NO_INTERFACE;
			}
			return this;
		};

		contact.firstName = aObjectItem.firstName;
		contact.lastName = aObjectItem.lastName; // not nillable
		contact.email = aObjectItem.primaryEmail;

		//there are no getter setter for folowing properties in tb3
		if (this.abVersion != 3) {

			contact.uid = aWillBeCreated ? null: aObjectItem.anniversaryDay;

			contact.aka = aObjectItem.nickName;
			contact.email2 = aObjectItem.secondEmail;

			contact.workPhone = aObjectItem.workPhone;
			contact.homePhone = aObjectItem.homePhone;
			contact.workFax = aObjectItem.faxNumber;
			contact.pagerPhone = aObjectItem.pagerNumber;
			contact.mobilePhone = aObjectItem.cellularNumber;

			// TODO smart merge
			contact.homeStreet = aObjectItem.homeAddress + " " + aObjectItem.homeAddress2;
			contact.homeZipCode = aObjectItem.homeZipCode;
			contact.homeCity = aObjectItem.homeCity;
			contact.homeState = aObjectItem.homeState;
			contact.homeExpressPostal = null;
			contact.homeCountry = aObjectItem.homeCountry;
			contact.homeWebPage = aObjectItem.webPage2;

			contact.title = aObjectItem.jobTitle;
			contact.company = aObjectItem.company;
			contact.service = aObjectItem.department;

			// TODO smart merge
			contact.workStreet = aObjectItem.workAddress + " " + aObjectItem.workAddress2;
			contact.workZipCode = aObjectItem.workZipCode;
			contact.workCity = aObjectItem.workCity;
			contact.workState = aObjectItem.workState;
			contact.workExpressPostal = null;
			contact.workCountry = aObjectItem.workCountry;
			contact.workWebPage = aObjectItem.webPage1;

			contact.comment = aObjectItem.notes;
			if (aObjectItem.custom1.indexOf('X-OBM-CALURI:') == 0)
				contact.caluri = aObjectItem.custom1.substr(13);
			else
				contact.caluri = '';
		} else {

			contact.uid = aWillBeCreated ? null: aObjectItem.getProperty("ObmId", null);

			contact.aka = aObjectItem.getProperty("NickName", null);
			contact.email2 = aObjectItem.getProperty("SecondEmail", null);

			contact.workPhone = aObjectItem.getProperty("WorkPhone", null);
			contact.homePhone = aObjectItem.getProperty("HomePhone", null);
			contact.workFax = aObjectItem.getProperty("FaxNumber", null);
			contact.pagerPhone = aObjectItem.getProperty("PagerNumber", null);
			contact.mobilePhone = aObjectItem.getProperty("CellularNumber", null);

			// TODO smart merge
			contact.homeStreet = aObjectItem.getProperty("HomeAddress", "") + " " + aObjectItem.getProperty("HomeAddress2", "");
			contact.homeZipCode = aObjectItem.getProperty("HomeZipCode", null);
			contact.homeCity = aObjectItem.getProperty("HomeCity", null);
			contact.homeState = aObjectItem.getProperty("HomeState", null);
			contact.homeExpressPostal = null;
			contact.homeCountry = aObjectItem.getProperty("HomeCountry", null);
			contact.homeWebPage = aObjectItem.getProperty("WebPage2", null);

			contact.title = aObjectItem.getProperty("JobTitle", null);
			contact.company = aObjectItem.getProperty("Company", null);
			contact.service = aObjectItem.getProperty("Department", null);

			// TODO smart merge
			contact.workStreet = aObjectItem.getProperty("WorkAddress", "") + " " + aObjectItem.getProperty("WorkAddress2", "");
			contact.workZipCode = aObjectItem.getProperty("WorkZipCode", null);
			contact.workCity = aObjectItem.getProperty("WorkCity", null);
			contact.workState = aObjectItem.getProperty("WorkState", null);
			contact.workExpressPostal = null;
			contact.workCountry = aObjectItem.getProperty("WorkCountry", null);
			contact.workWebPage = aObjectItem.getProperty("WebPage1", null);

			contact.comment = aObjectItem.getProperty("Notes", null);
			var custom1 = aObjectItem.getProperty("Custom1", null);
			if (custom1.indexOf('X-OBM-CALURI:') == 0)
				contact.caluri = custom1.substr(13);
			else
				contact.caluri = '';
		}
		return contact;
	},
	_computeDisplayName: function(aFirstName, aLastName) {
		return ((aFirstName && aFirstName != "") ? (aFirstName + " " + aLastName) : aLastName);
	},
	_updateLocalContact: function(localContact, aNewContact) {
		//compute displayName
		if (!localContact.displayName || localContact.displayName == "") {
			localContact.displayName = this._computeDisplayName(aNewContact.firstName, aNewContact.lastName);
		} else {
			if (localContact.displayName == this._computeDisplayName(localContact.firstName, localContact.lastName)) {
				localContact.displayName = this._computeDisplayName(aNewContact.firstName, aNewContact.lastName);
			}
		}

		localContact.firstName = aNewContact.firstName;
		localContact.lastName = aNewContact.lastName;
		localContact.primaryEmail = aNewContact.email;

		//there are no getter setter for folowing properties in tb3
		if (this.abVersion != 3) {

			localContact.anniversaryDay = aNewContact.uid;

			localContact.nickName = aNewContact.aka;
			localContact.secondEmail = aNewContact.email2;

			localContact.workPhone = aNewContact.workPhone;
			this.logger.debug("localContact.workPhone:" + aNewContact.workPhone, false, false);
			localContact.homePhone = aNewContact.homePhone;
			localContact.faxNumber = aNewContact.workFax;
			localContact.cellularNumber = aNewContact.mobilePhone;
			localContact.pagerNumber = aNewContact.pagerPhone;

			// TODO smart split
			localContact.homeAddress = aNewContact.homeStreet;
			localContact.workAddress2 = null;
			localContact.homeZipCode = aNewContact.homeZipCode;
			localContact.homeCity = aNewContact.homeCity;
			localContact.homeState = aNewContact.homeState;
			localContact.homeCountry = aNewContact.homeCountry;
			localContact.webPage2 = aNewContact.homeWebPage;

			localContact.jobTitle = aNewContact.title;
			localContact.department = aNewContact.service;
			localContact.company = aNewContact.company;

			// TODO smart split
			localContact.workAddress = aNewContact.workStreet;
			localContact.workAddress2 = null;
			localContact.workZipCode = aNewContact.workZipCode;
			localContact.workCity = aNewContact.workCity;
			localContact.workState = aNewContact.workState;
			localContact.workCountry = aNewContact.workCountry;
			localContact.webPage1 = aNewContact.workWebPage;

			localContact.notes = aNewContact.comment;
			if (aNewContact.caluri) {
				localContact.custom1 = 'X-OBM-CALURI:'+aNewContact.caluri;
			}
		} else {

			localContact.setProperty("ObmId", aNewContact.uid);
			localContact.setProperty("NickName", aNewContact.aka);

			localContact.setProperty("SecondEmail", aNewContact.email2);

			localContact.setProperty("WorkPhone", aNewContact.workPhone);
			localContact.setProperty("HomePhone", aNewContact.homePhone);
			localContact.setProperty("FaxNumber", aNewContact.workFax);
			localContact.setProperty("CellularNumber", aNewContact.mobilePhone);
			localContact.setProperty("PagerNumber", aNewContact.pagerPhone);

			// TODO smart split
			localContact.setProperty("HomeAddress", aNewContact.homeStreet);
			localContact.setProperty("WorkAddress2", null);
			localContact.setProperty("HomeZipCode", aNewContact.homeZipCode);
			localContact.setProperty("HomeCity", aNewContact.homeCity);
			localContact.setProperty("HomeState", aNewContact.homeState);
			localContact.setProperty("HomeCountry", aNewContact.homeCountry);
			localContact.setProperty("WebPage2", aNewContact.homeWebPage);

			localContact.setProperty("JobTitle", aNewContact.title);
			localContact.setProperty("Department", aNewContact.service);
			localContact.setProperty("Company", aNewContact.company);

			// TODO smart split
			localContact.setProperty("WorkAddress", aNewContact.workStreet);
			localContact.setProperty("WorkAddress2", null);
			localContact.setProperty("WorkZipCode", aNewContact.workZipCode);
			localContact.setProperty("WorkCity", aNewContact.workCity);
			localContact.setProperty("WorkState", aNewContact.workState);
			localContact.setProperty("WorkCountry", aNewContact.workCountry);
			localContact.setProperty("WebPage1", aNewContact.workWebPage);

			localContact.setProperty("Notes", aNewContact.comment);
			if (aNewContact.caluri) {
				localContact.setProperty('Custom1', 'X-OBM-CALURI:'+aNewContact.caluri);
			}
		}

		return localContact;
	},

	//
	// Resolve conflicts following override rule
	//
	_resolveConflicts: function(abName) {
		this.logger.debug("_resolveConflicts()", false, false);
		var serverPriority = (UtilsOBM.getCharPref("extensions.obm.override", "server") == "server");
		var remoteListHasChanged = false;
		for (var i = 0; i < this._remoteDiff[abName].removed.length; i++) {
			var contactId = "" + this._remoteDiff[abName].removed[i];
			var rLocalIdx = this._localDiff[abName].removed.indexOf(contactId);
			var uLocalIdx = this._localDiff[abName].updated.indexOf(contactId);
			if (rLocalIdx != -1) {
				// A conflict
				this._remoteDiff[abName].removed.splice(i--, 1);
				this._localDiff[abName].removed.splice(rLocalIdx, 1);
				UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory.toString(), abName, Ci.nsIAbCard.toString(), "removed", contactId);
				remoteListHasChanged = true;
			}
			if (uLocalIdx != -1) {
				// A conflict
				if (serverPriority) {
					this._localDiff[abName].updated.splice(uLocalIdx, 1);
					UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory.toString(), abName, Ci.nsIAbCard.toString(), "updated", contactId);
				} else {
					if (!remoteListHasChanged) this._remoteDiff[abName].removed.splice(i--, 1);
				}
			}
		}
		for (var i = 0; i < this._remoteDiff[abName].updated.length; i++) {
			remoteListHasChanged = false;
			var contactId = "" + this._remoteDiff[abName].updated[i].uid;
			var rLocalIdx = this._localDiff[abName].removed.indexOf(contactId);
			var uLocalIdx = this._localDiff[abName].updated.indexOf(contactId);
			if (uLocalIdx != -1) {
				// A conflict
				if (serverPriority) {
					this._localDiff[abName].updated.splice(uLocalIdx, 1);
					UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory.toString(), abName, Ci.nsIAbCard.toString(), "updated", contactId);
				} else {
					this._remoteDiff[abName].updated.splice(i--, 1);
					remoteListHasChanged = true;
				}
			}
			if (rLocalIdx != -1) {
				// A conflict
				if (serverPriority) {
					this._localDiff[abName].removed.splice(uLocalIdx, 1);
					UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory.toString(), abName, Ci.nsIAbCard.toString(), "removed", contactId);
				} else {
					if (!remoteListHasChanged) this._remoteDiff[abName].updated.splice(i--, 1);
				}
			}
		}

	},

	//
	// Query address book local modifications
	//
	_getLocalDiff: function(abName) {
		this.logger.debug("_getLocalDiff()", false, false);
		var changeTypes = ["added", "updated", "removed"];
		this._localDiff[abName] = {};
		for each(var type in changeTypes) {
			var pref = UtilsOBM.getCharPref("extensions.obm.changes." + Ci.nsIAbDirectory + "." + abName + "." + Ci.nsIAbCard + "." + type, "");
			var changes = (pref != "" ? pref.split(",") : []);
			this._localDiff[abName][type] = [];
			for (var i = 0; i < changes.length; i++) {
				if (changes[i] == null || changes[i] == "") {
					this.logger.error("_getItemsChanges id is null ("+"id:" + changes[i] + ", " + abName + ", " + type + ")", true, false);
				} else {
					this._localDiff[abName][type].push(changes[i]);
				}
			}
		}
	},
	_ensureValidResponse: function(aResCode, aMethodName, aResult, aValidMethodName) {
		if (aResCode != Cr.NS_OK) {
			var msgError = "unknown";
			if (aResult instanceof Ci.nsISupportsString) msgError = aResult;

			if (msgError == "No token available") {
				var username = {};
				var password = {};
				if (UtilsOBM.getOBMCredentials(username, password)) {
					UtilsOBM.setOBMCredentials(username.value, null);
					UtilsOBM.setTemporaryPassword(null);
				}
			}
			var printError = this._printableError(aMethodName, msgError);
			msgError = aMethodName + ": Reported error: " + msgError + "(" + aResCode + ")";
			this.logger.error(msgError, true, false);

			if (this.callerListener) {
				if (this._isCriticalMethod(aMethodName)) {
					this.callerListener.onStop(aResCode, printError);
				} else {
					this.callerListener.onError(printError);
				}
			}
			return false;
		}

		if (aValidMethodName && aMethodName != aValidMethodName) {
			var msgError = "Method " + aMethodName + " unexpected. Expected: " + aValidMethodName;
			this.logger.error(msgError, true, false);
			this.callerListener.onError(msgError);
			return false;
		}
		return true;
	},

	//
	// Get address book id from its name
	//
	_getABId: function(abName) {
		return (abName == "users" ? Ci.obmIRBook.USERS: Ci.obmIRBook.CONTACTS);
	},

	_onLogged: function(aResCode, aMethodName, aResult, aData) {
		this.callerListener.onOperationDone(true, null);
		this.logger.debug("_onLogged()", false, false);

		this.ABProxy.removeListener();
		if (!this._ensureValidResponse(aResCode, aMethodName, aResult, "doLogin")) {
			return;
		};
		this.callerListener.onNewOperation("getAddressBookSync", null);
		var self = this;
		this._getSyncOperationCount = 1;

		this._addListener(function(aResCode, aMethodName, aResult, aData) {
			self._onGetAddressBookSync(aResCode, aMethodName, aResult, aData)
		});

		this._localOperations = [];
		this._localOperations[0] = ["contacts", 0, 1,
			function(abName, aIdx, aTotal) {
				var lastSync = self._getLastSync();
				self.logger.debug("_sendNextLocalOperation()", false, false);
				self.callerListener.onProgress(aIdx, aTotal, "", "contacts");
				var data = new Context();
				data.isTheFirstPass = true;
				self.ABProxy.getAddressBookSync(lastSync, data);
			}];

		this._localOperationsIdx = 0;
		this._sendNextLocalOperation();
	},
	_onGetAddressBookSync: function(aResCode, aMethodName, aResult, aData) {
		this.logger.debug("_onGetAddressBookSync()", false, false);
		if (!this._ensureValidResponse(aResCode, aMethodName, aResult, "getAddressBookSync")) {
			return;
		};

		var data = aData.wrappedJSObject;
		var syncResponse = aResult.QueryInterface(Ci.obmIRAddressBookChangesResponse);

		this._addressbooks = {};
		this._addressbooks["removed"] = syncResponse.getRemovedAddressBooksIds({});
		this._addressbooks["updated"] = syncResponse.getUpdatedAddressBooks({});

		for each(var addressbookId in this._addressbooks["removed"]) {
			try {
				this.obmAbManager.removeAddressBook(addressbookId);
			} catch(e) {
				//Probably already removed
				this.logger.debug("addressbook " + addressbookId + " has been asked to be removed, but does not exist.", false, false, false);
			}
		}

		for each(var addressbook in this._addressbooks["updated"]) {
			// TODO. OBM_NEW_AD_MANAGER. use the new API  obmIAddressBooksManager
			this._ensureAddressBookIsPresent(addressbook);
		}
		// TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager::getAddressBookUids instead
		this.abNames = this._getAddressBookNames();

		for each(var abId in this.abNames) {
			this._remoteDiff[abId] = {};
			this._remoteDiff[abId]["updated"] = [];
			this._remoteDiff[abId]["removed"] = [];
			if (this._getLastSync() != "0") {
				/* Not possible to know where the contact comes from for the moment, so adding all removed contacts */
				this._remoteDiff[abId]["removed"] = syncResponse.getRemovedContactIds({});
			}
		}

		var localDiff = {};
		for (var i = 0; i < this.abNames.length; i++) {
			this._getLocalDiff(this.abNames[i]);
		}

		var locallyRemovedAddressBooksPref = UtilsOBM.getCharPref("extensions.obm.changes." + Ci.nsIAbDirectory + ".list." + Ci.nsIAbDirectory + ".removed", "");
		this._localDiff["unsubscribed"] = (locallyRemovedAddressBooksPref != "" ? locallyRemovedAddressBooksPref.split(",") : []);

		for each(var contact in syncResponse.getUpdatedContacts({})) {
			var length = this._remoteDiff[contact.addressbookid]["updated"].length;
			this._remoteDiff[contact.addressbookid]["updated"][length] = contact;
		}

		for each(var addressbook in this._addressbooks["updated"]) {
			if (data.isTheFirstPass) {
				this._resolveConflicts(addressbook.uid);
			}
		}

		this._setLastSync(syncResponse.lastsync);

		this._getSyncOperationCount--;
		this.logger.debug("obmRAddressBookSync.js: __syncAddressBooks: _setLastSync", false, false);

		if (this._getSyncOperationCount <= 0) {
			var self = this;
			this.ABProxy.removeListener();
			this.callerListener.onOperationDone(true, null);
			this._mergeRemoteDiff(data.isTheFirstPass);
		} else {
			this._sendNextLocalOperation();
		}
	},
	_mergeRemoteDiff: function(isTheFirstPass) {
		this.logger.debug("_mergeRemoteDiff()", false, false);

		var mergeActionCount = 0;
		var mergeAction = [];
		for each(var abName in this.abNames) {
			mergeAction[abName] = {};
			mergeAction[abName].updatedCount = this._remoteDiff[abName].updated.length;
			mergeAction[abName].removedCount = this._remoteDiff[abName].removed.length;
			mergeActionCount += this._remoteDiff[abName].updated.length;
			mergeActionCount += this._remoteDiff[abName].removed.length;
		}
		this.callerListener.onNewOperation("mergeRemoteDiff", mergeActionCount);
		this.logger.debug("Action count: " + mergeActionCount, false, false);

		var uIdx = 0;
		var rIdx = 0;
		// TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to retrieve the list
		// of addressbooks to add/update/remove a local contact.
		for each(var abName in this.abNames) {
			var addrDb = null;
			var directory = null;
			if (this.abVersion != 3) {
				addrDb = this._getDatabaseFromABName(abName);
			} else {
				directory = this._getDirectoryFromABName(abName);
			}

			for each(var updatedItem in this._remoteDiff[abName].updated) {
				try {
					// TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to get a local contact
					var localContact = this._getLocalContact(abName, updatedItem.uid);
					this.logger.debug(abName + "> WILL UPDATE: " + updatedItem.firstName, false, false);
					this._updateLocalContact(localContact, updatedItem);
					// TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to update a local contact
					if (this.abVersion != 3) {
						addrDb.editCard(localContact, false, null);
					} else {
						directory.modifyCard(localContact);
					}
				} catch(e) {
					try {
						if (this.abVersion != 3) {
							var localContact = {};
						} else {
							var localContact = Cc["@mozilla.org/addressbook/cardproperty;1"].createInstance(Ci.nsIAbCard);
						}
						this._updateLocalContact(localContact, updatedItem);
						this.logger.debug(abName + "> WILL CREATE: " + localContact.displayName, false, false);
						// TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to create a local contact
						if (this.abVersion != 3) {
							addrDb.createNewCardAndAddToDB(localContact, false);
						} else {
							directory.addCard(localContact);
						}
					} catch(e) {
						//this.logger.error("ERROR (ignored): " + e + "\n", true, false);
					}
				}
				if (this.abVersion != 3) {
					addrDb.commit(Ci.nsAddrDBCommitType.kSessionCommit);
				}
				this.callerListener.onProgress(uIdx++, mergeAction[abName].updatedCount, abName, "updated");
			}
			for each(var removedItem in this._remoteDiff[abName].removed) {
				this.logger.debug(abName + "> WILL REMOVE: " + removedItem, false, false);
				try {
					// TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to remove a local contact
					var contact = this._getLocalContact(abName, removedItem);
					if (this.abVersion != 3) {
						addrDb.deleteCard(contact, false, null);
						addrDb.commit(Ci.nsAddrDBCommitType.kSessionCommit);
					} else {
						var cardsToDelete = Cc["@mozilla.org/array;1"].createInstance(Ci.nsIMutableArray);
						cardsToDelete.appendElement(contact, false);
						directory.deleteCards(cardsToDelete);
					}
				} catch(e) {
					//this.logger.error("ERROR (ignored): " + e + "\n", true, false);
				}
				this.callerListener.onProgress(rIdx++, mergeAction[abName].removedCount, abName, "removed");
			}
			uIdx = 0;
			rIdx = 0;
		}

		this.callerListener.onOperationDone(true, null);
		if (isTheFirstPass) {
			this._sendLocalDiff();
		} else {
			this.callerListener.onStop(Cr.NS_OK, "done");
		}
	},
	_onSentLocalDiff: function(aResCode, aMethodName, aResult, aData) {
		this.logger.debug("_onSentLocalDiff()", false, false);
		this._localDiffOperationCount--;
		if (!this._ensureValidResponse(aResCode, aMethodName, aResult, null)) {
			// continue
		} else {
			switch (aMethodName) {
			case "createContact":
			case "createContactInBook":
			case "modifyContact":
				aData = aData.wrappedJSObject;
				var updatedContact = aResult.QueryInterface(Ci.obmIRContact);
				// TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to get a local contact
				var localContact = this._getLocalContact(aData.abName, aData.id);
				this._updateLocalContact(localContact, updatedContact);
				// TODO. OBM_NEW_AD_MANAGER. use obmIAddressBooksManager to update a local contact
				if (this.abVersion != 3) {
					var addrDb = this._getDatabaseFromABName(aData.abName);
					addrDb.editCard(localContact, false, null);
					addrDb.commit(Ci.nsAddrDBCommitType.kSessionCommit);
				} else {
					var directory = this._getDirectoryFromABName(aData.abName);
					directory.modifyCard(localContact);
				}
				break;
			case "removeContact":
			case "unsubscribeBook":
				break;
			default:
			};
		}
		if (this._localDiffOperationCount <= 0) {
			this.logger.debug("Operations done", false, false);
			this.callerListener.onOperationDone(true, null);
			var self = this;
			this.ABProxy.removeListener();
			var timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
			this.logger.debug("timer init", false, false);
			this.callerListener.onNewOperation("wait-one-seconde", null);
			timer.initWithCallback({
				notify: function(timer) {
					self.logger.debug("ok, time is out", false, false);
					self._fireSecondPass();
				}
			},
			1100, timer.TYPE_ONE_SHOT);
			this.timer = timer;
		} else {
			this._sendNextLocalOperation();
		}
	},
	_fireSecondPass: function() {
		this.logger.debug("_fireSecondPass()", false, false);
		this._getSyncOperationCount = 1;
		var self = this;
		this.callerListener.onOperationDone(true, null);
		this.callerListener.onNewOperation("getAddressBookSync", null);
		this._addListener(function(aResCode, aMethodName, aResult, aData) {
			self._onGetAddressBookSync(aResCode, aMethodName, aResult, aData)
		});

		this._localOperations = [];
		this._localOperations[0] = ["contacts", 0, 1,
			function(abName, aIdx, aTotal) {
				var lastSync = self._getLastSync();
				self.logger.debug("_sendNextLocalOperation()", false, false);
				self.callerListener.onProgress(aIdx, aTotal, "", "contacts");
				var data = new Context();
				data.isTheFirstPass = false;
				self.ABProxy.getAddressBookSync(lastSync, data);
			}];

		this._localOperationsIdx = 0;
		this._sendNextLocalOperation(false);
	},
	_sendLocalDiff: function() {
		this.logger.debug("_sendLocalDiff()", false, false);
		var queriesCount = 0;
		for (var i = 0; i < this.abNames.length; i++) {
			var abName = this.abNames[i];
			if (abName == "users") {
				// TODO revert change, ...
				continue;
			}
			queriesCount += this._localDiff[abName].removed.length;
			queriesCount += this._localDiff[abName].updated.length;
			queriesCount += this._localDiff[abName].added.length;
		}
		queriesCount += this._localDiff["unsubscribed"].length;

		this.callerListener.onNewOperation("sendLocalDiff", queriesCount);

		this._localDiffOperationCount = queriesCount;
		var self = this;
		this._addListener(function(aResCode, aMethodName, aResult, aData) {
			self._onSentLocalDiff(aResCode, aMethodName, aResult, aData)
		});

		if (queriesCount == 0) {
			this._onSentLocalDiff(Cr.NS_OK, null, null, null);
			return;
		}

		var idx = 0;
		this._localOperations = [];
		for (var i = 0; i < this.abNames.length; i++) {
			var abName = this.abNames[i];
			for (var j = 0; j < this._localDiff[abName].removed.length; j++) {
				this._localOperations[idx++] = [abName, j, this._localDiff[abName].removed.length,
				function(abName, aIdx, aTotal) {
					self.callerListener.onProgress(aIdx, aTotal, "removed", abName);
					self.ABProxy.removeContact(self._getABId(abName), self._localDiff[abName].removed[aIdx], null);
					UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory, abName, Ci.nsIAbCard, "removed", self._localDiff[abName].removed[aIdx]);
				}];
			}
			for (var j = 0; j < this._localDiff[abName].added.length; j++) {
				this._localOperations[idx++] = [abName, j, this._localDiff[abName].added.length,
				function(abName, aIdx, aTotal) {
					self.callerListener.onProgress(aIdx, aTotal, "added", abName);
					var ctx = new Context();
					ctx.abName = abName;
					ctx.id = self._localDiff[abName].added[aIdx];
					self.ABProxy.createContactInBook(abName, self._createContactObjFromId(abName, ctx.id, true), ctx);
					UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory, abName, Ci.nsIAbCard, "added", ctx.id);
				}];
			}
			for (var j = 0; j < this._localDiff[abName].updated.length; j++) {
				this._localOperations[idx++] = [abName, j, this._localDiff[abName].updated.length,
				function(abName, aIdx, aTotal) {
					self.callerListener.onProgress(aIdx, aTotal, "updated", abName);
					var ctx = new Context();
					ctx.abName = abName;
					ctx.id = self._localDiff[abName].updated[aIdx];
					self.ABProxy.modifyContact(self._getABId(abName), self._createContactObjFromId(abName, ctx.id, false), ctx);
					UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory, abName, Ci.nsIAbCard, "updated", ctx.id);
				}];
			}
		}
		for (var j = 0; j < this._localDiff.unsubscribed.length; j++) {
			this._localOperations[idx++] = [this._localDiff.unsubscribed[j], j, this._localDiff.unsubscribed.length,
			function(abName, aIdx, aTotal) {
				self.callerListener.onProgress(aIdx, aTotal, "unsubscribed", abName);
				var ctx = new Context();
				ctx.abName = abName;
				ctx.id = self._localDiff.unsubscribed[aIdx];
				self.ABProxy.unsubscribeBook(ctx.id, ctx);
				UtilsOBM.removeItemFromChangesList(Ci.nsIAbDirectory, "list", Ci.nsIAbDirectory, "removed", ctx.id);
			}];
		}
		this._localOperationsIdx = 0;
		this._sendNextLocalOperation();
	},
	_sendNextLocalOperation: function() {
		var idx = this._localOperationsIdx++;
		var self = this;
		this._localOperations[idx][3](this._localOperations[idx][0], this._localOperations[idx][1], this._localOperations[idx][2]);
	},
	// DEPRECATED. OBM_NEW_AD_MANAGER. use the new API  obmIAddressBooksManager::ensureAddressBookExists
	_ensureAddressBookIsPresent: function(addressbook) {
		this.logger.debug("_ensureAddressBookIsPresent()", false, false);
		var addressBookName = addressbook.uid;
		var dirName = addressbook.name;
		/* Get configured name for "contacts" and "collected_contacts" */
		dirName = UtilsOBM.getCharPref("extensions.obm.addressbooks." + dirName, dirName);
		if (addressbook.ownerDisplayName != null && addressbook.ownerDisplayName != "") {
			dirName += " (" + addressbook.ownerDisplayName + ")";
		}
		if (this.abVersion == 3) {
			var addressBookManager = Components.classes["@mozilla.org/abmanager;1"].getService(Components.interfaces.nsIAbManager);
			var abURI = "moz-abmdbdirectory://" + addressBookName + ".mab";
			var dirPrefId = "ldap_2.servers." + addressBookName;
			var iter = addressBookManager.directories;
			var found = false;
			while (iter.hasMoreElements()) {
				var data = iter.getNext();
				if (data instanceof Ci.nsIAbDirectory) {
					if (data.dirPrefId == dirPrefId) {
						found = true;
						/* Update name if necessary */
						if (data.dirName != dirName) {
							data.dirName = dirName;
						}
					}
				}
			}
			if (!found) {
				addressBookManager.newAddressBook(dirName, abURI, 2, dirPrefId);
				//this._setLastSync(addressBookName, "0");
				this._addAddressBookNames(addressBookName);
			}
		} else {
			var abURI = "moz-abmdbdirectory://alia-obm-" + addressBookName + ".mab";
			var directory = RDFService.GetResource(abURI).QueryInterface(Ci.nsIAbDirectory);
			var children;
			try {
				children = directory.childCards;
				// Ok, address books are here, skiping creation
				/* Update name if necessary */
				if (directory.dirName != dirName) {
					directory.dirName = dirName;
				}
			} catch(e) {
				if (e.name == "NS_ERROR_FILE_NOT_FOUND") {
					//this._setLastSync(addressBookName, "0");
					var directoryProperties = Cc["@mozilla.org/addressbook/properties;1"].getService(Ci.nsIAbDirectoryProperties);
					directoryProperties.description = dirName;
					// What 2 is ?
					directoryProperties.dirType = 2;
					directoryProperties.fileName = abURI.split('/').pop();
					var addressBookService = Cc["@mozilla.org/addressbook;1"].getService(Ci.nsIAddressBook);
					addressBookService.newAddressBook(directoryProperties);
					this._addAddressBookNames(addressBookName);
				} else {
					this.logger.error(e, true, false);
					this.callerListener.onError(e);
					throw e;
				}
			}

		}
	},

	/*
	 * obmIRAddressBookSync
	 */

	mergeAddressBook: function(aListener) {
		this.logger.debug("mergeAddressBook()", false, false);
		this.callerListener = aListener;
		this.callerListener.onStart();

		//
		// Remote jobs
		//
		this.callerListener.onNewOperation("logUserIn", null);
		var login = {};
		var password = {};
		this._getOBMCredentials(login, password);

		try {
			var PortUrl = UtilsOBM.getCharPref("extensions.obm.server"); // +
			// "/AddressBook";
		} catch(e) {
			var error = "NO_SERVER_INFO";
			this.logger.error(error, true, false);
			this.callerListener.onError(error);
			throw (error);
		}

		this.ABProxy = Cc["@obm.org/obm-sync-rest/addressbookproxy;1"].createInstance(Ci.obmIRAddressBookProxy);
		this.ABProxy.PortUrl = PortUrl;

		var self = this;
		this._addListener(function(aResCode, aMethodName, aResult, aData) {
			self._onLogged(aResCode, aMethodName, aResult, aData)
		});
		this.ABProxy.doLogin(login.value, password.value, UtilsOBM.getOrigin(), null);
	}
};

// module
var ObmRAddressBookSyncModule = {
	mCID: Components.ID("{68214b68-8421-480c-8cd3-bd0cf70a50c4}"),
	mContractID: "@obm.org/obm-sync-rest/addressbooksync;1",
	registerSelf: function(compMgr, fileSpec, location, type) {
		compMgr = compMgr.QueryInterface(Ci.nsIComponentRegistrar);
		compMgr.registerFactoryLocation(this.mCID, "OBM Address Book Sync", this.mContractID, fileSpec, location, type);
	},
	getClassObject: function(compMgr, cid, iid) {

		if (!cid.equals(this.mCID)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		if (!iid.equals(Ci.nsIFactory)) {
			throw Cr.NS_ERROR_NOT_IMPLEMENTED;
		}
		if (!this._utilsLoaded) {
			var scriptLoader = Cc["@mozilla.org/moz/jssubscript-loader;1"].createInstance(Ci.mozIJSSubScriptLoader);
			scriptLoader.loadSubScript("chrome://obm-extension/content/calendar/UtilsOBM.js");
			this._utilsLoaded = true;
		}

		return this.mFactory;
	},
	mFactory: {
		createInstance: function(outer, iid) {
			if (outer != null) {
				throw Cr.NS_ERROR_NO_AGGREGATION;
			}
			return (new ObmRAddressBookSync()).QueryInterface(iid);
		}
	},

	canUnload: function(compMgr) {
		return true;
	}
};

function NSGetModule(compMgr, fileSpec) {
	return ObmRAddressBookSyncModule;
}
