/* ***** BEGIN LICENSE BLOCK *****
 * 
 * Copyright (c) 1997-2008 Aliasource - Groupe LINAGORA
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  http://www.obm.org/                                              
 * 
 * ***** END LICENSE BLOCK ***** */
/* Base Synchronizer */

function Base_Synchronizer() {
	//Init logger
	this.logger = Components.classes["@obm.org/calendar/logger;1"].createInstance(Components.interfaces.obmILogger);
	var writer = Components.classes["@obm.org/calendar/writer;1?type=unixfile"].getService(Components.interfaces.obmILoggerWriter);
	this.logger.setLevelWriter(this.logger.LOAD_LEVEL, writer);
	this.logger.setHeader("Base_Synchronizer: ");
	this.logger.debug("Base_Synchronizer()", false, false);
	this.abVersion = Components.classes["@mozilla.org/abmanager;1"] ? 3 : 2;
	this.globalData = Components.classes["@obm.org/globaldata;1"].getService();
}

Base_Synchronizer.prototype = {

	get mSynchronizing() {
		return this.globalData.wrappedJSObject.addrSynchronizing; // this value is set by ObmSyncManager
	},

	shutdown: function() {
		this.mProxy.unsubscribreToObserverService();
	},
	notifyItemAdded: function(aContainer, aItem) {
		this.logger.debug("notifyItemAdded", false, false);
		if (this.mSynchronizing) {
			this.logger.debug("mSynchronizing", false, false);
			return;
		}

		aContainer = aContainer.QueryInterface(this.mSupportedContainerType);

		if (!this.isContainerManagedByOBM(aContainer)) {
			this.logger.debug("! isContainerManagedByOBM", false, false);
			return;
		}

		var containerName = this.getContainerOBMName(aContainer);

		var itemID = this.getItemID(aItem);
		var newId = false;
		if (!itemID) {
			newId = true;
			// assign an ID to the new created item if necessary
			// (i.e. address book contact)
			itemID = "" + Math.round(Math.random() * 1000) + "-" + Math.round(Math.random() * 1000);

			this.logger.debug("itemID:" + itemID, false, false);

			if (this.mSupportedContainerType == Components.interfaces.nsIAbDirectory && this.abVersion != 3) {
				this.logger.debug("nsIAbDirectory", false, false);
				var windowManager = Components.classes['@mozilla.org/appshell/window-mediator;1'].getService(Components.interfaces.nsIWindowMediator);
				var addressBookWindow = windowManager.getMostRecentWindow("mail:addressbook");
				if (addressBookWindow) {
					// XXX selecting the same address book doesn't reload the data!
					// re-select the current address book to refresh the view after synchronization
					var view = addressBookWindow.document.getElementById("dirTree").view;
					view.selection.select(0); //invalidateSelection();
				}
			}
		}

		UtilsOBM.addItemToChangesList(this.mSupportedContainerType.toString(), containerName, this.mSupportedItemType.toString(), "added", itemID);
		if (newId) {
			this.setItemID(aContainer, aItem, itemID);
			//in tb3 will raise onPropertyChanged of addressbook listenner
			//so call after pushing in added list
		}
	},

	notifyItemModified: function(aContainer, aItem) {
		this.logger.debug("notifyItemModified", false, false);
		if (this.mSynchronizing) {
			this.logger.debug("mSynchronizing", false, false);
			return;
		}

		aContainer = aContainer.QueryInterface(this.mSupportedContainerType);
		aItem = aItem.QueryInterface(this.mSupportedItemType);

		if (!this.isContainerManagedByOBM(aContainer)) {
			return;
		}

		var containerName = this.getContainerOBMName(aContainer);

		var itemID = this.getItemID(aItem);
		this.logger.debug("itemID:" + itemID, false, false);

		UtilsOBM.addItemToChangesList(this.mSupportedContainerType.toString(), containerName, this.mSupportedItemType.toString(), "updated", itemID);
	},

	notifyItemParticipationChanged: function(aContainer, aItem) {
		this.logger.debug("notifyItemParticipationChanged", false, false);
		if (this.mSynchronizing) {
			this.logger.debug("mSynchronizing", false, false);
			return;
		}

		aContainer = aContainer.QueryInterface(this.mSupportedContainerType);
		aItem = aItem.QueryInterface(this.mSupportedItemType);

		if (!this.isContainerManagedByOBM(aContainer)) {
			return;
		}

		var containerName = this.getContainerOBMName(aContainer);

		var itemID = this.getItemID(aItem);
		this.logger.debug("itemID:" + itemID, false, false);

		UtilsOBM.addItemToChangesList(this.mSupportedContainerType.toString(), containerName, this.mSupportedItemType.toString(), "participationChanged", itemID);
	},

	notifyItemDeleted: function(aContainer, aItem) {
		this.logger.debug("notifyItemDeleted", false, false);
		if (this.mSynchronizing) {
			this.logger.debug("mSynchronizing", false, false);
			return;
		}

		aContainer = aContainer.QueryInterface(this.mSupportedContainerType);
		aItem = aItem.QueryInterface(this.mSupportedItemType);

		if (!this.isContainerManagedByOBM(aContainer)) {
			return;
		}

		var containerName = this.getContainerOBMName(aContainer);

		var itemID = this.getItemID(aItem);
		if (!itemID) {
			return;
		}

		UtilsOBM.addItemToChangesList(this.mSupportedContainerType.toString(), containerName, this.mSupportedItemType.toString(), "removed", itemID);
	},

	notifyContainerDeleted: function(aContainer) {
		this.logger.debug("notifyContainerDeleted", false, false);
		if (this.mSynchronizing) {
			this.logger.debug("mSynchronizing", false, false);
			return;
		}

		aContainer = aContainer.QueryInterface(this.mSupportedContainerType);

		if (!this.isContainerManagedByOBM(aContainer)) {
			this.logger.debug("! isContainerManagedByOBM", false, false);
			return;
		}

		var containerName = this.getContainerOBMName(aContainer);

		UtilsOBM.addItemToChangesList(this.mSupportedContainerType.toString(), "list", this.mSupportedContainerType.toString(), "removed", this.getContainerOBMName(aContainer));
		UtilsOBM.deletePreferenceBranch("extensions.obm.changes." + this.mSupportedContainerType.toString() + "." + this.getContainerOBMName(aContainer));
		// Delete addressbook from addressbook list
		// TODO. OBM_NEW_AD_MANAGER. properly define API instead...
		var name = this.getContainerOBMName(aContainer);
		var namesPref = UtilsOBM.getCharPref("extensions.obm.addressbooks", null);
		var names = namesPref ? namesPref.split(",") : [];
		var newNames = [];
		for (var i=0; i<names.length; i++) {
			if (names[i] != name) {
				newNames.push(names[i]);
			}
		}
		UtilsOBM.setCharPref("extensions.obm.addressbooks", newNames.join(","));
		this.logger.debug("Update extensions.obm.addressbooks to " + newNames.join(","));

	},

	/* PRIVATE FUNCTIONS */

	fillItemsHashMap: function(aContainer, aItemsIDArray) {
		var itemsList = new HashMap();
		for each(var itemID in aItemsIDArray) {
			itemsList.put(itemID, this.getItemFromContainer(aContainer, itemID));
		}
		return itemsList;
	},

	isContainerManagedByOBM: function(aContainer) {
		return this.getContainerOBMName(aContainer) != null;
	},

	_removeItemFromChangesList: function(aContainer, aChangeType, aItemID) {
		UtilsOBM.resetAllItemsChangesLists(this.mSupportedContainerType.toString(), this.getContainerOBMName(aContainer), this.mSupportedItemType.toString(), aChangeType, aItemID);
	}
};
